#!/usr/bin/env python3

# This script handles AP Linux OSUpgrades
# The script randomizes the startup time to ensure that not all machines attempt simultaneous upgrades since PS
# servers won't be able to handle the load
#
# 1. The script reads the existing Environment.ini brought in by syncdata service and determines the image label
#    that the customer has chosen to use for this particular machine function/environment
# 2. The script downloads wim.ini using the psfileclient.py library and checks what the current version label is of the
#    image label used by Environment.ini
# 3. The script compares the version from wim.ini to the current running system OsVersion at /opt/ap/OsVersion
# 4. If the versions match, it does nothing
# 5. If the versions do not match, an OSUpgrade is required
# 6. The script initiates an OSUpgrade repair to DM for this machine
# 7. The script also posts a property to DM to let DM know that it has acquired this new label
# 8. The machine is then repaired via OSUpgrade by DM and PS servers and comes up with the new OS

import os
import sys
import csv
import io
import logging
import socket
import random
import time

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apcommon import iniparser
from apsdk import apenv
from apsdk import dmclient
from apsdk import psfileclient

# The default umask after AzSecPack remediation is 077 which is too restrictive in cases where we want
# services to make their files visible publicly by default without having to set permissions on them after
# we create them.  Setting the umask to 022 at the beginning of a script makes sure the default behavior is
# default Linux permissions.  If permissions need to be specific, we can set them using python functions.
os.umask(0o022)


class OSUpgrade(object):
    WIM_INI_NAME = "wim.ini"
    WIM_DESTINATION = "/opt/ap/wim.ini"
    ENVIRONMENT_INI = "/ap/data/AutopilotData/PE/Environment.ini"
    OSVERSION_FILE = "/opt/ap/OsVersion"
    WATCHDOG_CSV_DIR = "/ap/data/watchdogoutput/csv_sources"
    WATCHDOG_TMP_FILE = "osupgrade.tmp"
    WATCHDOG_CSV_FILE = "osupgrade.csv"
    REPAIRS_TO_IGNORE = ["BIOSUpgrade"]

    # Setup logging
    def log_setup(self) -> bool:
        try:
            logger = logging.getLogger()
            logger.setLevel(logging.INFO)
            formatter = logging.Formatter('%(asctime)s [%(process)d]: %(message)s', '%b %d %H:%M:%S')

            handler_stdout = logging.StreamHandler(sys.stdout)
            handler_stdout.setFormatter(formatter)
            handler_stdout.addFilter(type('', (logging.Filter,), {'filter': staticmethod(lambda r: r.levelno <= logging.INFO)}))
            logger.addHandler(handler_stdout)

            handler_stderr = logging.StreamHandler(sys.stderr)
            handler_stderr.setFormatter(formatter)
            handler_stderr.setLevel(logging.WARNING)
            logger.addHandler(handler_stderr)
        except Exception as e:
            print("osupgrade.log_setup: Failed to setup logging. Error: {0}".format(e), file=sys.stderr)
            return False

        return True

    # Get current hostname
    def get_hostname(self) -> tuple:
        logging.debug("osupgrade:get_hostname: Getting hostname")
        try:
            hostname = socket.gethostname().lower()
        except Exception as e:
            logging.error("osupgrade.get_hostname: Failed to get hostname via socket.gethostname(). Error: {0}".format(e))
            return "", False

        logging.debug("osupgrade.get_hostname: Hostname: {0}".format(hostname))
        return hostname, True

    # Get the current machine function
    def get_machine_function(self) -> tuple:
        logging.debug("osupgrade:get_machine_function: Getting machine function")
        try:
            machine_function = apenv.get_machine_function().lower()
        except Exception as e:
            logging.error("osupgrade.get_machine_function: Failed to get machine function name. Error: {0}".format(e))
            return "", False

        logging.debug("osupgrade.get_machine_function: Machine Function: {0}".format(machine_function))
        return machine_function, True

    # Get the current machine environment
    def get_environment(self) -> tuple:
        logging.debug("osupgrade:get_environment: Getting environment")
        try:
            environment = apenv.get_environment_name().lower()
        except Exception as e:
            logging.error("osupgrade.get_environment: Failed to get environment name. Error: {0}".format(e))
            return "", False

        logging.debug("osupgrade.get_environment: Environment: {0}".format(environment))
        return environment, True

    # Get the cluster name
    def get_cluster(self) -> tuple:
        logging.debug("osupgrade:get_cluster: Getting cluster")
        try:
            cluster = apenv.get_cluster_name().lower()
        except Exception as e:
            logging.error("osupgrade.get_cluster: Failed to get cluster name. Error: {0}".format(e))
            return "", False

        logging.debug("osupgrade.get_cluster: Cluster: {0}".format(cluster))
        return cluster, True

    # Check to make sure file exists
    def check_file_exists(self, filename: str) -> bool:
        logging.debug("osupgrade.check_file_exists: Checking file exists: {0}".format(filename))
        if not os.path.exists(filename):
            logging.error("osupgrade.check_file_exists: {0} file not found".format(filename))
            return False

        return True

    # Initialize INIParser class to read ini files
    def initialize_iniparser(self, ini_file: str) -> tuple:
        logging.debug("osupgrade.initialize_iniparser: initializing INIParser for: {0}".format(ini_file))
        try:
            iniparser_client = iniparser.INIParser(ini_file, case_sensitive=False)
        except Exception as e:
            logging.error("osupgrade.initialize_iniparser: Could not initialize INIParser")
            logging.error("osupgrade.initialize_iniparser: Error: {0}".format(e))
            return "", False

        if not isinstance(iniparser_client, iniparser.INIParser):
            return "", False

        try:
            iniparser_client.read()
        except Exception as e:
            logging.error("osupgrade.initialize_iniparser: Could not read inifile")
            logging.error("osupgrade.initialize_iniparser: Error: {0}".format(e))
            return "", False

        return iniparser_client, True

    # Get the image tag associated with the machine_function
    def get_environment_image_tag(self, machine_function: str, iniparser_environment_ini_client: iniparser.INIParser) -> tuple:
        section = "MaintenanceService"
        parameter = "OsImage"
        mf_parameter = "{0}_{1}".format(parameter, machine_function)
        tag_value = None

        logging.debug("osupgrade.get_environment_image_tag: Looking for MF specific image tag")
        try:
            tag_value = iniparser_environment_ini_client.get(section, mf_parameter.lower())
        except Exception as e:
            logging.error("osupgrade.get_environment_image_tag: Could not read inifile Section: {0}, Parameter: {1}".format(section, mf_parameter))
            logging.error("osupgrade.get_environment_image_tag: Warning: {0}".format(e))

        # If we found an MF specific image tag this is the one used and we return it right away
        # If not, we keep looking
        if tag_value and ("ubuntu" in tag_value.lower() or "linux" in tag_value.lower()):
            logging.debug("osupgrade.get_environment_image_tag: Found MF specific image tag: {0}".format(tag_value))
            return tag_value, True

        logging.debug("osupgrade.get_environment_image_tag: Looking for global default image tag")
        try:
            tag_value = iniparser_environment_ini_client.get(section, parameter.lower())
        except Exception as e:
            logging.error("osupgrade.get_environment_image_tag: Could not read inifile Section: {0}, Parameter: {1}".format(section, parameter))
            logging.error("osupgrade.get_environment_image_tag: Warning: {0}".format(e))

        # If we find the global image tag for the whole environment we return it since we couldn't find an MF specific one
        if tag_value and ("ubuntu" in tag_value.lower() or "linux" in tag_value.lower()):
            logging.debug("osupgrade.get_environment_image_tag: Found global image tag: {0}".format(tag_value))
            return tag_value, True

        # No OsImage tag found; can't determine upgrade
        logging.error("osupgrade.get_environment_image_tag: Error: Could not obtain image tag from Environment.ini")
        return "", False

    # Get the image tag value from the wim file to compare against existing machine osversion
    def get_wim_image_tag_value(self, tag_name: str, iniparser_wim_ini_client: iniparser.INIParser) -> tuple:
        section = "Images"
        tag_value = None

        # Customer did not define tag name in Environment.ini so the default image tag for Linux architecture is used
        # from wim.ini
        if not tag_name:
            section = "WIM"
            tag_name = "Linux.retail.amd64"

        logging.debug("osupgrade.get_wim_image_tag_value: Looking for wim.ini tag value for tag: {0}".format(tag_name))
        try:
            tag_value = iniparser_wim_ini_client.get(section, tag_name)
        except Exception as e:
            logging.error("osupgrade.get_wim_image_tag_value: Could not read inifile Section: {0}, Parameter: {1}".format(section, tag_name))
            logging.error("osupgrade.get_wim_image_tag_value: Warning: {0}".format(e))

        # If we found the tag and it has a value return the value
        if tag_value:
            logging.debug("osupgrade.get_wim_image_tag_value: Found tag: {0} Value: {1}".format(tag_name, tag_value))
            return tag_value, True

        return "", False

    # Initialize PS server file client
    def initialize_ps_file_client(self) -> tuple:
        try:
            ps_file_client = psfileclient.PSFileClient()
        except Exception as e:
            logging.error("osupgrade.initialize_ps_file_client: Could not initialize PSFileClient")
            logging.error("osupgrade.initialize_ps_file_client: Error: {0}".format(e))
            return "", False

        if not isinstance(ps_file_client, psfileclient.PSFileClient):
            return "", False

        return ps_file_client, True

    # Parse the wim image tag value into an osversion
    def get_osversion_from_image_tag(self, tag_value: str) -> tuple:
        # Now we have to split the image tag to get the actual OSVersion
        # Example: Ubuntu1604.Debug=Ubuntu1604-amd64-180929Debug-d641b77a2f8a9cc4ee08366ce5ead3c0.wim,18.09.29Ubuntu1604.Debug
        try:
            split_tag = tag_value.strip().split(",")
            osversion = split_tag[1]
        except Exception as e:
            logging.error("osupgrade.get_osversion_from_image_tag: Could not get OSVersion from image tag: {0}".format(tag_value))
            logging.error("osupgrade.get_osversion_from_image_tag: Error: {0}".format(e))
            return "", False

        logging.debug("osupgrade.get_osversion_from_image_tag: Found WIM OSVersion: {0}".format(osversion))
        return osversion, True

    # Get the current machine OSVersion
    def get_machine_osversion(self, osversion_file: str=OSVERSION_FILE) -> tuple:
        logging.debug("osupgrade:get_osversion: Getting current machine OsVersion from: {0}".format(osversion_file))

        if os.path.isfile(osversion_file):
            try:
                with open(osversion_file, 'r') as f:
                    first_line = f.readline().strip()
                    f.close()
            except Exception as e:
                logging.error("osupgrade:get_osversion: Exception: {0}".format(e))
                return "", False

            # The only supported OS right now is Ubuntu so we match that to make sure it's valid
            if "Ubuntu" in first_line:
                osversion = first_line
            else:
                logging.error("osupgrade:get_osversion: OsVersion does not match Ubuntu")
                return "", False
        else:
            logging.error("osupgrade:get_osversion: {0} does not exist.".format(osversion_file))
            return "", False

        if osversion:
            logging.debug("osupgrade.get_osversion_from_image_tag: Found current machine OSVersion: {0}".format(osversion))
            return osversion, True
        else:
            logging.error("osupgrade:get_osversion: OsVersion is empty!")
            return "", False

    # Get DM Gateway URL
    def get_dm_gateway_url(self) -> tuple:
        logging.debug("osupgrade:get_dm_gateway_url: Getting DM gateway URL")
        try:
            dm_gateway_url = apenv.get_secure_dm_gateway_url().lower()
        except Exception as e:
            logging.error("osupgrade.get_dm_gateway_url: Failed to get DM gateway URL. Error: {0}".format(e))
            return "", False

        logging.debug("osupgrade.get_secure_dm_gateway_url: DM gateway URL: {0}".format(dm_gateway_url))
        return dm_gateway_url, True

    # Get dm_client object
    def create_dm_client(self, dm_gateway_url: str) -> tuple:
        logging.debug("osupgrade:create_dm_client: Initialize DM client")
        try:
            dm_client = dmclient.DMClient(dm_gateway_url)
        except Exception as e:
            logging.error("osupgrade.create_dm_client: Failed to initialize DM client. Error: {0}".format(e))
            return "", False

        logging.debug("osupgrade.create_dm_client: DM client initialized")
        return dm_client, True

    # Get the repair list from DM
    def get_repairs_list_from_dm(self, machine_function: str, dm_client: dmclient) -> tuple:
        command = "GetRepairList -fields RepairActionState,MachineName,RequestType,RequestTime -f " + machine_function
        logging.debug("osupgrade:get_repairs_list_from_dm: Command: {0}".format(command))
        try:
            response = dm_client.send_request(command, '')
        except Exception as e:
            logging.error("osupgrade:get_repairs_list_from_dm: Exception: {0}".format(e))
            return "", False

        if not response.text:
            logging.error("osupgrade:get_repairs_list_from_dm: No response.text returned")
            return "", False

        logging.debug("osupgrade:get_repairs_list_from_dm: DM response: {0}".format(response))
        logging.debug("osupgrade:get_repairs_list_from_dm: DM response text: \n{0}".format(response.text))

        return response.text, True

    # Verify that we got a valid DM response
    def check_dm_response(self, repairs_list: str) -> bool:
        if repairs_list.lower().startswith("no response"):
            logging.debug("osupgrade:check_dm_response Got 'No Response' response from DM")
            return False

        return True

    # Check if a DM Repair is in progress for this host
    def check_repair_in_progress(self, repairs_list: str, hostname: str) -> tuple:
        # Repair list format from DM
        #   #Fields:RepairActionState,MachineName,RequestType,RequestTime
        #   Proceed,CO4PHY080090121,OSUpgrade,2018-10-10 19:31:52.058
        #   #EOF
        repair = ""
        lines = repairs_list.splitlines()
        reader = csv.reader(lines)
        for row in reader:
            if len(row) == 4 and row[1].strip() == hostname:
                repair = row[2].strip()
                break

        if repair and (repair not in self.REPAIRS_TO_IGNORE):
            logging.debug("osupgrade:check_repair_in_progress: Found Repair: {0} on Host: {1}".format(repair, hostname))
            return repair, True
        else:
            logging.debug("osupgrade:check_repair_in_progress: Found no repair on Host: {0}".format(hostname))
            return "", False

    # Initiate OSUpgrade repair request to DM
    def initiate_osupgrade_repair_request_to_dm(self, hostname: str, environment: str, dm_client: dmclient) -> bool:
        command = "Repair -s {0} -w {1}  -t OSUpgrade -c \"AP Linux OSUpgrade\" -repairErrorFree -force -retryCount 2".format(hostname, environment)
        logging.debug("osupgrade:initiate_osupgrade_repair_request_to_dm: Command: {0}".format(command))

        try:
            response = dm_client.send_request(command, '')
        except Exception as e:
            logging.error("osupgrade:initiate_osupgrade_repair_request_to_dm: Exception: {0}".format(e))
            return False

        if not response.text:
            logging.error("osupgrade:initiate_osupgrade_repair_request_to_dm: No response.text returned")
            return False

        logging.debug("osupgrade:initiate_osupgrade_repair_request_to_dm: DM response: {0}".format(response))
        logging.debug("osupgrade:initiate_osupgrade_repair_request_to_dm: DM response text: \n{0}".format(response.text))

        return True

    # Create DM report out of the information returned
    def build_dm_report(self, hostname: str, osversion: str, key: str="SyncOSImage") -> tuple:
        dm_output = io.StringIO()

        status = "Ok"
        value = "Got images:{0}".format(osversion)
        line = [hostname.upper(), status, key, value]

        try:
            logging.debug(",".join(line) + '\r\n')
            dm_output.write(",".join(line) + '\r\n')
        except Exception as e:
            logging.error("osupgrade.build_dm_report: Could not build DM report from OsVersion")
            logging.error("osupgrade.build_dm_report: Error: {0}".format(e))
            return "", False

        return dm_output, True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self, dm_output: io.StringIO) -> bool:
        logging.debug("osupgrade.verify_dm_report: Verifying DM report")
        report = dm_output.getvalue().split("\r\n")
        logging.debug("osupgrade.verify_dm_report: {0}".format(report))\

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,SyncOSImage,Got images:
            if not len(results) == 4:
                logging.error("osupgrade.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self, dm_output: io.StringIO, watchdog_csv_dir: str=WATCHDOG_CSV_DIR, watchdog_tmp_file: str=WATCHDOG_TMP_FILE) -> bool:
        if not os.path.isdir(watchdog_csv_dir):
            logging.error("osupgrade.write_dm_report_to_disk: Directory: {0} not found".format(watchdog_csv_dir))
            return False
        else:
            tmp_file_path = os.path.join(watchdog_csv_dir, watchdog_tmp_file)
            logging.debug("osupgrade.write_dm_report_to_disk: Writing file: {0} to disk".format(tmp_file_path))

            try:
                with open(tmp_file_path, "w") as text_file:
                    text_file.write(dm_output.getvalue())
            except Exception as e:
                logging.error("osupgrade.write_dm_report_to_disk: Could not write tmp file {0} to disk".format(tmp_file_path))
                logging.error("osupgrade.write_dm_report_to_disk: Error: {0}".format(e))
                return False

        logging.debug("osupgrade.write_dm_report_to_disk: Successfully wrote tmp file: {0}".format(tmp_file_path))
        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self, dm_output: io.StringIO, watchdog_csv_dir: str=WATCHDOG_CSV_DIR, watchdog_tmp_file: str=WATCHDOG_TMP_FILE) -> bool:
        tmp_file_path = os.path.join(watchdog_csv_dir, watchdog_tmp_file)
        logging.debug("osupgrade.verify_write_dm_report_to_disk: Verifying file: {0} to disk".format(tmp_file_path))

        if os.path.exists(tmp_file_path):
            try:
                with open(tmp_file_path, 'r') as content_file:
                    content = content_file.read()

            except Exception as e:
                logging.error("osupgrade.verify_write_dm_report_to_disk: Could not read tmp file {0} from disk".format(tmp_file_path))
                logging.error("osupgrade.verify_write_dm_report_to_disk: Error: {0}".format(e))
                return False

            content = content.split("\n")
            dm_output = dm_output.getvalue().split("\r\n")

            if list(filter(None, content)) == list(filter(None, dm_output)):
                logging.debug("osupgrade.verify_write_dm_report_to_disk: File: {0} matches DM report written".format(tmp_file_path))
                return True
            else:
                logging.error("osupgrade.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                return False

        else:
            logging.error("osupgrade.verify_write_dm_report_to_disk: File: {0} does not exist!".format(tmp_file_path))
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self, watchdog_csv_dir: str=WATCHDOG_CSV_DIR, watchdog_tmp_file: str=WATCHDOG_TMP_FILE, watchdog_csv_file: str=WATCHDOG_CSV_DIR) -> bool:
        tmp_file_path = os.path.join(watchdog_csv_dir, watchdog_tmp_file)
        csv_file_path = os.path.join(watchdog_csv_dir, watchdog_csv_file)

        try:
            logging.debug("osupgrade.atomically_rename_dm_report_on_disk Atomically renaming: {0} to {1}".format(tmp_file_path, csv_file_path))
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            logging.error("osupgrade.atomically_rename_dm_report_on_disk: Could not atomically rename: {0} to {1}".format(tmp_file_path, csv_file_path))
            logging.error("osupgrade.atomically_rename_dm_report_on_disk: Error: {0}".format(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self, dm_output: io.StringIO) -> bool:
        output = "\n".join(dm_output.getvalue().split("\r\n"))
        dm_output.close()

        try:
            logging.debug("DM Report:")
            logging.debug(output)
        except Exception as e:
            logging.error("osupgrade.get_environment_image_tag: Could not output DM report")
            logging.error("osupgrade.print_dm_report: Error: {0}".format(e))
            return False

        return True

    def main(self):
        # Setup logging
        if not self.log_setup():
            print("osupgrade.main: Failed to get setup logging (main.log_setup)", file=sys.stderr)
            return False

        logging.info("osupgrade.main: Starting osupgrade")

        # Sleep between 4 hours and 23 hours
        # Randomizes the kickoff of OSUpgrades so that they don't all start at the same time
        # The service is restarted every 24 hours by systemd so we have time to spare
        # We give 4 hours at the start of the service in case of a long rollout to an environment.  We don't want
        # machines going into OSUpgrade repair mode while the rollouts are happening for any new deployment
        seconds_to_sleep = random.randint(14400, 82800)
        logging.info("osupgrade.main: Sleeping for {0} seconds in order to randomize OSUpgrade start times".format(seconds_to_sleep))
        time.sleep(seconds_to_sleep)

        # Get current hostname
        hostname, success = self.get_hostname()
        if not success or not hostname:
            logging.error("osupgrade.main: Failed to get hostname (main.get_hostname)")
            sys.exit(1)

        # Get the current machine function
        machine_function, success = self.get_machine_function()
        if not success or not machine_function:
            logging.error("osupgrade.main: Failed to get machine_function (main.get_machine_function)")
            sys.exit(1)

        # Get the current machine environment
        environment, success = self.get_environment()
        if not success or not environment:
            logging.error("osupgrade.main: Failed to get environment (main.get_environment)")
            sys.exit(1)

        # Get the cluster name
        cluster, success = self.get_cluster()
        if not success or not cluster:
            logging.error("osupgrade.main: Failed to get cluster (main.get_cluster)")
            sys.exit(1)

        # Check to make sure Environment.ini file is synced and exists; otherwise this is moot
        if not self.check_file_exists(self.ENVIRONMENT_INI):
            logging.info("osupgrade.main: {0} not synced yet (main.check_environment_ini_exists)".format(self.ENVIRONMENT_INI))
            logging.info("osupgrade.main: Will try again next cycle (main.check_environment_ini_exists)")
            sys.exit(0)

        # Initialize INIParser class to read ini files
        iniparser_environment_ini_client, success = self.initialize_iniparser(self.ENVIRONMENT_INI)
        if not success:
            logging.error("osupgrade.main: Failed to initialize INIParser object for {0} (main.initialize_iniparser)".format(self.ENVIRONMENT_INI))
            sys.exit(1)

        # Get the image tag associated with the machine_function
        image_tag, success = self.get_environment_image_tag(machine_function, iniparser_environment_ini_client)
        if not success or not image_tag:
            logging.error("osupgrade.main: Failed to get image_tag from Environment.ini (main.get_environment_image_tag)")
            logging.info("osupgrade.main: Customer did not configure a custom image_tag for Environment.ini Linux machines")
            logging.info("osupgrade.main: Will use default image tag: Linux.retail.amd64")

        # Initialize PS server file client
        ps_file_client, success = self.initialize_ps_file_client()
        if not success:
            logging.error("osupgrade.main: Failed to initialize PSFileClient object (main.ps_file_client.initialize_ps_file_client)")
            sys.exit(1)

        # Download WIM file from PS server
        if not ps_file_client.process_psserver_file_action(self.WIM_INI_NAME, self.WIM_DESTINATION, False, False):
            logging.error("osupgrade.main: Failed to get wim.ini file (main.ps_file_client.process_psserver_file_action)")
            sys.exit(1)

        # Initialize INIParser class to read ini files
        iniparser_wim_ini_client, success = self.initialize_iniparser(self.WIM_DESTINATION)
        if not success:
            logging.error("osupgrade.main: Failed to initialize INIParser object for {0} (main.initialize_iniparser)".format(self.WIM_DESTINATION))
            sys.exit(1)

        # Get the image tag value from the wim file to compare against existing machine osversion
        wim_tag_value, success = self.get_wim_image_tag_value(image_tag, iniparser_wim_ini_client)
        if not success or not wim_tag_value:
            logging.error("osupgrade.main: Failed to get wim.ini image tag value (main.get_wim_image_tag_value)")
            sys.exit(1)

        # Parse the wim image tag value into an osversion
        wim_osversion, success = self.get_osversion_from_image_tag(wim_tag_value)
        if not success or not wim_osversion:
            logging.error("osupgrade.main: Failed to get osversion from image tag value (main.get_osversion_from_image_tag)")
            sys.exit(1)

        # Get the current machine OsVersion
        machine_osversion, success = self.get_machine_osversion()
        if not success or not machine_osversion:
            logging.error("osupgrade.main: Failed to get current machine osversion from {0} (main.get_machine_osversion)".format(self.OSVERSION_FILE))
            sys.exit(1)

        # If the OsVersion has changed; take action
        if wim_osversion != machine_osversion:

            logging.info("osupgrade.main: OSUpgrade required from OS: {0} to New OS: {1}".format(machine_osversion, wim_osversion))

            # Get DM Gateway URL
            dm_gateway_url, success = self.get_dm_gateway_url()
            if not success or not dm_gateway_url:
                logging.error("osupgrade.main: Failed to get DM gateway URL")
                sys.exit(1)

            # Get dm_client object
            dm_client, success = self.create_dm_client(dm_gateway_url)
            if not success:
                logging.error("osupgrade.main: Failed to initialize DM client")
                sys.exit(1)

            # Get the repair list from DM
            repairs_list, success = self.get_repairs_list_from_dm(machine_function, dm_client)
            if not success or not repairs_list:
                logging.error("osupgrade.main: Failed to get DM repair list for MF: {0}".format(machine_function))
                sys.exit(1)

            # Verify that we got a valid DM response
            if not self.check_dm_response(repairs_list):
                logging.error("osupgrade.main: DM returned invalid response: {0}".format(repairs_list))
                sys.exit(1)

            # Check if a DM Repair is in progress for this host
            # If there's already a repair in progress we exit
            repair, success = self.check_repair_in_progress(repairs_list, hostname)
            if success and repair:
                logging.info("osupgrade.main: Repair: {0} already in progress. Will try again next cycle".format(repair))
                sys.exit(0)

            # Initiate OSUpgrade repair request to DM
            if not self.initiate_osupgrade_repair_request_to_dm(hostname, environment, dm_client):
                logging.error("osupgrade.main: Failed to initiate an OSUpgrade repair request to DM")
                sys.exit(1)

            logging.info("osupgrade.main: OSUpgrade request successfully sent to DM")

            #
            #
            # Once BIOSUpgrade scenario works on Linux this can be used
            # Until then, we can only have one active repair by DM service at a time which means BIOSUpgrade
            # breaks our ability to OSUpgrade machines
            # Until this is resolved on the AP side, etc. we will have to initiate the OSUpgrade repair directly from
            # inside this watchdog to get around this blocker
            # The following block of code gets the right property to DM but won't take any action until
            # we resolve the outstanding problem.
            # For now, we leave this in for compatibility with DM Repair Service but it's not used to initialize the
            # repair
            #
            #

            # Create DM report out of the information returned
            dm_output, success = self.build_dm_report(hostname, wim_osversion)
            if not success:
                logging.error("osupgrade.main: Failed to build DM report (main.build_dm_report)")
                sys.exit(1)

            # Make sure the DM report is properly formatted
            if not self.verify_dm_report(dm_output):
                logging.error("osupgrade.main: Failed to verify DM report (main.verify_dm_report)")
                sys.exit(1)

            if not self.write_dm_report_to_disk(dm_output, self.WATCHDOG_CSV_DIR, self.WATCHDOG_TMP_FILE):
                logging.error("osupgrade.main: Failed to write temp file to disk (main.write_dm_report_to_disk)")
                sys.exit(1)

            if not self.verify_write_dm_report_to_disk(dm_output, self.WATCHDOG_CSV_DIR, self.WATCHDOG_TMP_FILE):
                logging.error("osupgrade.main: Failed to verify temp file on disk (main.verify_write_dm_report_to_disk)")
                sys.exit(1)

            if not self.atomically_rename_dm_report_on_disk(self.WATCHDOG_CSV_DIR, self.WATCHDOG_TMP_FILE, self.WATCHDOG_CSV_FILE):
                logging.error("osupgrade.main: Failed to atomically rename file on disk (main.atomically_rename_dm_report_on_disk)")
                sys.exit(1)

            if not self.print_dm_report(dm_output):
                sys.exit(1)

        else:
            logging.info("osupgrade.main: No OSUpgrade required at this time")

        logging.info("osupgrade.main: Completed osupgrade")
        sys.exit(0)


if __name__ == '__main__':
    OSUpgrade().main()
